@@ -1,59 +1,70 @@
 /* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
 /* GeoIPCity.c
  *
- * Copyright (C) 2002 MaxMind.com
+ * Copyright (C) 2006 MaxMind LLC
  *
  * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public
+ * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public
+ * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #include <GeoIP.h>
+#include <GeoIP_internal.h>
 #include <GeoIPCity.h>
+#ifndef WIN32
 #include <netdb.h>
+#include <netinet/in.h> /* For ntohl */
+#else
+#include <windows.h>
+#include <winsock.h>
+#endif
+#include <sys/types.h> /* For uint32_t */
+#ifdef HAVE_STDINT_H
+#include <stdint.h>     /* For uint32_t */
+#endif
 
+static
 const int FULL_RECORD_LENGTH = 50;
 
-GeoIPRecord * _get_record(GeoIP* gi, unsigned long ipnum) {
+static
+GeoIPRecord * _extract_record(GeoIP* gi, unsigned int seek_record, int *next_record_ptr) {
 	int record_pointer;
 	unsigned char *record_buf = NULL;
 	unsigned char *begin_record_buf = NULL;
 	GeoIPRecord * record;
 	int str_length = 0;
 	int j;
-        unsigned int seek_country;
 	double latitude = 0, longitude = 0;
 	int dmaarea_combo = 0;
-
-	if (gi->databaseType != GEOIP_CITY_EDITION_REV0 &&
-			gi->databaseType != GEOIP_CITY_EDITION_REV1) {
-		printf("Invalid database type %s, expected %s\n", GeoIPDBDescription[(int)gi->databaseType], GeoIPDBDescription[GEOIP_CITY_EDITION_REV1]);
-		return 0;
-	}
-
-	seek_country = _seek_country(gi, ipnum);
-	if (seek_country == gi->databaseSegments[0])		
+	int bytes_read = 0;
+	if (seek_record == gi->databaseSegments[0])		
 		return NULL;
 
 	record = malloc(sizeof(GeoIPRecord));
 	memset(record, 0, sizeof(GeoIPRecord));
 
-	record_pointer = seek_country + (2 * gi->record_length - 1) * gi->databaseSegments[0];
+	record_pointer = seek_record + (2 * gi->record_length - 1) * gi->databaseSegments[0];
 
 	if (gi->cache == NULL) {
 		fseek(gi->GeoIPDatabase, record_pointer, SEEK_SET);
 		begin_record_buf = record_buf = malloc(sizeof(char) * FULL_RECORD_LENGTH);
-		fread(record_buf, sizeof(char), FULL_RECORD_LENGTH, gi->GeoIPDatabase);
+		bytes_read = fread(record_buf, sizeof(char), FULL_RECORD_LENGTH, gi->GeoIPDatabase);
+		if (bytes_read == 0) {
+			/* eof or other error */
+			free(begin_record_buf);
+			free(record);
+			return NULL;
+		}
 	} else {
 		record_buf = gi->cache + (long)record_pointer;
 	}
@@ -118,35 +129,76 @@ GeoIPRecord * _get_record(GeoIP* gi, unsigned long ipnum) {
 	if (gi->cache == NULL)
 		free(begin_record_buf);
 
+	/* Used for GeoIP_next_record */
+	if (next_record_ptr != NULL)
+		*next_record_ptr = seek_record + record_buf - begin_record_buf + 3;
+
 	return record;
 }
 
+static
+GeoIPRecord * _get_record(GeoIP* gi, unsigned long ipnum) {
+	unsigned int seek_record;
+
+	if (gi->databaseType != GEOIP_CITY_EDITION_REV0 &&
+			gi->databaseType != GEOIP_CITY_EDITION_REV1) {
+		printf("Invalid database type %s, expected %s\n", GeoIPDBDescription[(int)gi->databaseType], GeoIPDBDescription[GEOIP_CITY_EDITION_REV1]);
+		return 0;
+	}
+
+	seek_record = _GeoIP_seek_record(gi, ipnum);
+	return _extract_record(gi, seek_record, NULL);
+}
+
+GeoIPRecord * GeoIP_record_by_ipnum (GeoIP* gi, unsigned long ipnum) {
+	return _get_record(gi, ipnum);
+}
+
 GeoIPRecord * GeoIP_record_by_addr (GeoIP* gi, const char *addr) {
 	unsigned long ipnum;
 	if (addr == NULL) {
 		return 0;
 	}
-	ipnum = _addr_to_num(addr);
+	ipnum = _GeoIP_addr_to_num(addr);
 	return _get_record(gi, ipnum);
 }
 
 GeoIPRecord * GeoIP_record_by_name (GeoIP* gi, const char *name) {
 	unsigned long ipnum;
-	struct hostent * host;
 	if (name == NULL) {
 		return 0;
 	}
-	ipnum = _addr_to_num(name);
-	if (ipnum == 0) {
-		host = gethostbyname(name);
-		if (host == NULL) {
-			return 0;
-		}
-		ipnum = _h_addr_to_num((unsigned char *) host->h_addr_list[0]);
-	}
+	ipnum = _GeoIP_lookupaddress(name);
 	return _get_record(gi, ipnum);
 }
 
+int GeoIP_record_id_by_addr (GeoIP* gi, const char *addr) {
+	unsigned long ipnum;
+	if (gi->databaseType != GEOIP_CITY_EDITION_REV0 &&
+			gi->databaseType != GEOIP_CITY_EDITION_REV1) {
+		printf("Invalid database type %s, expected %s\n", GeoIPDBDescription[(int)gi->databaseType], GeoIPDBDescription[GEOIP_CITY_EDITION_REV1]);
+		return 0;
+	}
+	if (addr == NULL) {
+		return 0;
+	}
+	ipnum = _GeoIP_addr_to_num(addr);
+	return _GeoIP_seek_record(gi, ipnum);
+}
+
+int GeoIP_init_record_iter (GeoIP* gi) {
+	return gi->databaseSegments[0] + 1;
+}
+
+int GeoIP_next_record (GeoIP* gi, GeoIPRecord **gir, int *record_iter) {
+	if (gi->cache != NULL) {
+		printf("GeoIP_next_record not supported in memory cache mode\n");
+		return 1;
+	}
+	*gir = _extract_record(gi, *record_iter, record_iter);
+	return 0;
+}
+
 void GeoIPRecord_delete (GeoIPRecord *gir) {
 	free(gir->region);
 	free(gir->city);
